from ocli import Base, Main, arg, flag, param
from ocli.extra import BasicLog, Counter, DryRunOpt, LogOpt


def filesizef(s):
    if not s and s != 0:
        return "-"
    for x in "bkMGTPEZY":
        if s < 1000:
            break
        s /= 1024.0
    return ("%.1f" % s).rstrip("0").rstrip(".") + x


class Counter2(Counter):
    def _format_value(self, value, key):
        # print(key, value)
        if key in ("size", "disk_size"):
            return filesizef(value)
        return str(value)


# @arg(append="paths", required="+")
# @arg("action", choices=("link", "stat", "unique_files", "debug"), required=True)
# @flag("dry_run", help="Test run")
# @flag("carry_on", help="Continue on file errors", default=None)
# @param(
#     "linker",
#     "l",
#     help="The linker to use",
#     choices=("os.link", "ln", "lns", "os.symlink"),
#     default="os.link",
# )
# class App(BasicLog, Main):
class App(LogOpt, DryRunOpt, Base):
    dry_run = False

    def options(self, opt):
        super().options(
            opt.arg(
                "action",
                choices=("link", "stat", "unique_files", "debug"),
                required=True,
            )
            .arg(append="paths", required="+")
            .flag("carry_on", help="Continue on file errors", default=None)
            .param(
                "linker",
                "l",
                help="The linker to use",
                choices=("os.link", "ln", "lns", "os.symlink"),
                default="os.link",
            )
        )

    def start(self, **kwargs):
        from logging import error, info
        from os import stat
        from stat import S_ISDIR

        from . import (Database, dump_db, get_linker, link_duplicates,
                       list_uniques, scan_dir)

        # print("start", self.__dict__)
        # print("action", self.action)
        # print("linker", self.paths)

        db = Database()
        tot = self.total = Counter2()
        carry_on = self.carry_on

        def statx(f):
            try:
                st = stat(f)
            except Exception:
                tot.file_err += 1
                if not carry_on:
                    raise
                from sys import exc_info

                error(exc_info()[1])
                return 0, 0, 0, 0, 0

            return st.st_mode, st.st_size, st.st_ino, st.st_dev, st.st_mtime

        for x in self.paths:
            mode, size, ino, dev, mtime = statx(x)
            if S_ISDIR(mode):
                scan_dir(x, db, statx)
            else:
                db.add(x, size, ino, dev, mtime)
        action = self.action

        try:
            if action == "debug":
                dump_db(db)
            elif action == "unique_files":
                list_uniques(db, self.total)
            elif action == "stat":
                link_duplicates(db, None, self.total, self.carry_on)
            elif action == "link":
                link_duplicates(
                    db,
                    get_linker(self.linker),
                    self.total,
                    self.carry_on,
                )
            else:
                raise RuntimeError(f"Unknown action {action}")
        finally:
            self.total and info("Total {}".format(self.total))
        return self.total


def main():
    return App().main()


(__name__ == "__main__") and App().main()



import unittest
from hashlib import md5
from tempfile import gettempdir


class Test(unittest.TestCase):
    def test_link(self):
        import string
        from os.path import join, split, exists
        from os import makedirs
        from shutil import rmtree

        tmp = join(gettempdir(), "dupln")

        data = dict(
            filter(
                (lambda _: isinstance(_[1], str)),
                ((k, getattr(string, k)) for k in dir(string) if k[0].isalpha()),
            )
        )

        # pprint.pprint(data)

        def test(sepins=[2, 5, 8]):
            exists(tmp) and rmtree(tmp)
            same_file = {}

            def put_bytes(b, path):
                h = md5()
                h.update(b)
                q = (len(b), h.hexdigest())
                if q in same_file:
                    same_file[q] += 1
                else:
                    same_file[q] = 1
                dir = split(path)[0]
                exists(dir) or makedirs(dir)
                with open(path, "wb") as o:
                    o.write(b)

            for k, v in data.items():
                b = v.encode("UTF-8")
                put_bytes(b[::-1], join(tmp, k))  # reversed
                for i in sepins:
                    if i < len(k):
                        k = k[:i] + "/" + k[i:]
                        put_bytes(b, join(tmp, k))

            return dict(
                same_file=sum(1 for count in same_file.values() if count > 1),
                unique_files=len(same_file),
                disk_size=sum(size_hash[0] for size_hash in same_file.keys()),
                size=sum(
                    size_hash[0] * count for size_hash, count in same_file.items()
                ),
                files=sum(same_file.values()),
                same_size=len(set(size_hash[0] for size_hash in same_file.keys())),
                linked=sum(count - 1 for count in same_file.values() if count > 1),
            )

        v = test()

        from .__main__ import App

        # Total disk_size 836b; files 26; inodes 26; same_size 8; size 836b;
        total = App().main(["cmd", "stat", tmp])
        self.assertEqual(total.disk_size, v["size"])
        self.assertEqual(total.files, v["files"])
        self.assertEqual(total.inodes, v["files"])
        self.assertEqual(total.same_size, v["same_size"])
        self.assertEqual(total.size, v["size"])
        # Total devices 1; disk_size 836b; files 26; inodes 26; size 836b; unique_size 8;
        total = App().main(["cmd", "unique_files", tmp])
        self.assertEqual(total.disk_size, v["size"])
        self.assertEqual(total.files, v["files"])
        self.assertEqual(total.inodes, v["files"])
        self.assertEqual(total.size, v["size"])
        self.assertEqual(total.unique_size, v["same_size"])
        # disk_size ; files ; inodes ; linked ; same_hash ; same_size ; size ;
        total = App().main(["cmd", "link", tmp])
        self.assertEqual(total.disk_size, v["disk_size"])
        self.assertEqual(total.files, v["files"])
        self.assertEqual(total.inodes, v["files"])
        self.assertEqual(total.linked, v["linked"])
        self.assertEqual(total.same_hash, v["same_file"])
        self.assertEqual(total.same_size, v["same_size"])
        self.assertEqual(total.size, v["size"])
        #
        total = App().main(["cmd", "unique_files", tmp])
        self.assertEqual(total.disk_size, v["disk_size"])
        self.assertEqual(total.files, v["files"])
        self.assertEqual(total.inodes, v["unique_files"])
        self.assertEqual(total.size, v["size"])
        self.assertEqual(total.unique_size, v["same_size"])
        #
        total = App().main(["cmd", "stat", tmp])
        self.assertEqual(total.disk_size, v["disk_size"])
        self.assertEqual(total.files, v["files"])
        self.assertEqual(total.same_size, v["same_size"])
        self.assertEqual(total.inodes, v["unique_files"])
        self.assertEqual(total.size, v["size"])



from hashlib import md5
from os.path import join, dirname, exists, samefile
from os import walk, unlink, chmod, rename, link, symlink, stat
from logging import info, error
from tempfile import mktemp
from stat import S_ISREG, S_IWUSR


class Database(dict):
    def add(self, path, size, ino, dev, mtime):

        # dev_map = self.get((dev, size))

        dev_map = self.get(dev)
        if dev_map is None:
            dev_map = self[dev] = {}

        size_map = dev_map.get(size)
        if size_map is None:
            size_map = dev_map[size] = {}

        ino_map = size_map.get(ino)
        if ino_map is None:
            # ino_map = size_map[ino] = {}
            ino_map = size_map[ino] = set()

        ino_map.add(path)

        # v = ino_map.get(path)
        # if v is None:
        #     ino_map[path] = 0


def calc_md5(src, block_size=131072):
    m = md5()
    with open(src, "rb") as h:
        b = h.read(block_size)
        while b:
            m.update(b)
            b = h.read(block_size)
    return m.hexdigest()


def file_sort_key(x):
    return stat(x).st_mtime


def link_duplicates(db, linker, tot, carry_on):
    if len(db) > 1:
        tot.devices = len(db)
    while db:
        dev, size_map = db.popitem()
        while size_map:
            size, ino_map = size_map.popitem()
            if len(ino_map) < 2 or size < 1:
                while ino_map:
                    ino, paths = ino_map.popitem()
                    n = len(paths)
                    if n > 1:
                        tot.same_ino += 1
                    tot.files += n
                    tot.inodes += 1
                    tot.size += n * size
                    tot.disk_size += size
                continue
            else:
                tot.same_size += 1
            md5_map = {} if linker else 0

            while ino_map:
                ino, paths = ino_map.popitem()
                n = len(paths)
                if n > 1:
                    tot.same_ino += 1
                tot.files += n
                tot.inodes += 1
                if md5_map == 0:
                    tot.size += size * n
                    tot.disk_size += size
                else:
                    src = paths.pop()
                    md5 = calc_md5(src)
                    files = md5_map.get(md5)
                    if files is None:
                        files = md5_map[md5] = set()
                    files.add(src)

            while md5_map:
                md5, files = md5_map.popitem()
                n = len(files)
                tot.size += size * n
                if n > 1:
                    tot.same_hash += 1
                    files = sorted(files, key=file_sort_key)
                    try:
                        n = link_dups(linker, files)
                    except Exception:
                        tot.link_err += 1
                        if not carry_on:
                            raise
                        from sys import exc_info

                        error(exc_info()[1])
                    else:
                        tot.linked += n
                        tot.disk_size += size
                else:  # n == 1
                    tot.uniq_hash += 1
                    tot.disk_size += size


def link_dups(linker, dups):
    src = n = 0
    while dups:
        dup = dups.pop()
        if src:
            assert dup != src
            assert exists(dup)
            tmp = mktemp(dir=dirname(dup))
            # rename dup to tmp
            rename(dup, tmp)
            assert exists(tmp)
            assert not exists(dup)
            info(" - %r - %r [%s]", dup, tmp, len(dups))
            try:  # link src to dup
                linker(src, dup)
            except OverflowError:
                # rename back on error
                exists(tmp) and rename(tmp, dup)
                src = dup
                info("\t! Too many links")
            except Exception:
                # rename back on error
                info("\t! Link failed")
                exists(tmp) and rename(tmp, dup)
                raise
            else:  # delete tmp
                n += 1
                assert exists(dup)
                assert samefile(src, dup)
                chmod(tmp, S_IWUSR)
                unlink(tmp)
        else:
            src = dup
            info("++ %r [%s]", src, len(dups))
    return n


def get_linker(use_linker):
    from subprocess import call

    if use_linker == "ln":

        def fun(src, dst):
            call(["ln", src, dst])

    elif use_linker == "lns":

        def fun(src, dst):
            call(["ln", "-s", src, dst])

    elif use_linker == "os.link":

        def fun(src, dst):
            link(src, dst)

    elif use_linker == "os.symlink":

        def fun(src, dst):
            symlink(src, dst)

    else:
        raise RuntimeError(f"Unknown linker {use_linker!r}")
    return fun


def list_uniques(db, tot):
    tot.devices = len(db)

    while db:
        dev, size_map = db.popitem()
        tot.unique_size = len(size_map)
        while size_map:
            size, ino_map = size_map.popitem()
            while ino_map:
                ino, paths = ino_map.popitem()
                n = len(paths)
                if n > 1:
                    tot.same_ino += 1
                tot.files += n
                tot.inodes += 1
                tot.size += n * size
                tot.disk_size += size
                path = paths.pop()
                print(path)


def dump_db(db):
    from sys import stdout

    if len(db) == 1:
        db = db.popitem()[1]
    try:
        from yaml import dump, safe_dump
    except ImportError:
        from json import dump, JSONEncoder

        class SetEncoder(JSONEncoder):
            def default(self, obj):
                if isinstance(obj, set):
                    return list(obj)
                return JSONEncoder.default(self, obj)

        dump(db, stdout, cls=SetEncoder, indent=4)

    else:
        return safe_dump(db, stdout, canonical=False, tags=False, indent=True)


def scan_dir(tree, db, statx):
    info("Scanning: %r", tree)
    for root, dirs, files in walk(tree):
        for name in files:
            f = join(root, name)
            (mode, size, ino, dev, mtime) = statx(f)
            S_ISREG(mode) and db.add(f, size, ino, dev, mtime)



from setuptools import setup, find_packages

setup(
    name="dupln",
    version="0.0.4",
    author="biojet1",
    author_email="biojet1@gmail.com",
    description="Hard link files with same content",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/biojet1/dupln",
    packages=find_packages(),
    entry_points={
        "console_scripts": ["dupln=dupln.__main__:main"],
    },
    install_requires=["ocli"],
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
)
